package traffic;

import java.net.*;
import java.util.*;
import traffic.object.*;

public class Simulator implements Constants {
  private final IO io;

  public Simulator(InetAddress kernelAddress, int kernelPort) {
    io = new IO(kernelAddress, kernelPort);
    io.sendConnect();
    io.receiveConnectOk();
    io.sendAcknowledge();
    setInitialPosition();
  }

  private void setInitialPosition() {
    for (int i = WORLD.movingObjectArray().length - 1;  i >= 0;  i --) {
      MovingObject mv = WORLD.movingObjectArray()[i];
      Lane lane;
      MotionlessObject mlpos = mv.motionlessPosition();
      if (mlpos instanceof Building) {
	lane = null;
      }
      else if (mlpos instanceof Node) {
	List ols = ((Node) mlpos).outLanes();
	lane = initialPosition(ols, 0);
 	lane = (lane instanceof Lane) ? lane : (Lane) Util.getAtRandom(ols);
      }
      else {
	if(ASSERT)Util.assert(mlpos instanceof Road, "unexpected motionlessPosition", mlpos);
	List lsh = ((Road) mlpos).lanesToHead();
	List lst = ((Road) mlpos).lanesToTail();
	Lane lh = initialPosition(lsh, mv.positionExtra());
	Lane lt = initialPosition(lst, mv.positionExtra());
	boolean hOrT = RANDOM.nextDouble() < lsh.size() / (lsh.size() + lst.size());
	if (lh instanceof Lane  &&  (lt == null  ||  hOrT))
	  lane = lh;
	else if (lt instanceof Lane)
	  lane = lt;
	else
	  lane = (Lane) Util.getAtRandom((hOrT) ? lsh : lst);
      }
      mv.setLane(lane);
    }
  }

  private Lane initialPosition(List candidates, double positionExtra) {
    Iterator it = candidates.iterator();
    while (it.hasNext()) {
      Lane lane = (Lane) it.next();
      if (!isThereOthers(lane, positionExtra))
	return lane;
    }
    return null;
  }

  private boolean isThereOthers(Lane lane, double positionExtra) {
    Iterator it = lane.mvObjOnSet().iterator();
    while (it.hasNext()) {
      MovingObject mv = (MovingObject) it.next();
      if (mv.positionExtra() == positionExtra)
	return true;
    }
    return false;
  }

  public void simulate() {
    System.out.println("start");
    while (true) {
      io.receiveCommands();
      System.out.println("time: " + WORLD.time());
      move();
      loadUnload();
      io.sendUpdate();
      io.receiveUpdate();
    }
  }

  private void move() {
    MovingObject[] mvObjs = WORLD.movingObjectArray();
    for (int i = mvObjs.length - 1;  i >= 0;  i --)
      mvObjs[i].initializeEveryCycle();

    MovingObject.resetNumReached();

    long startTime = System.currentTimeMillis();

    for (WORLD.setSec(0);  WORLD.sec() < 60;  WORLD.progressSec()) {
      if (MovingObject.numReached() == mvObjs.length
	  || System.currentTimeMillis() - startTime > CALCULATING_LIMIT_MILLI_SEC)
	break;

      if (USE_VIEWER) VIEWER.waitProgress();
      if (USE_VIEWER) VIEWER.setTime();

      for (int i = mvObjs.length - 1;  i >= 0;  i --) {
	MovingObject mv = mvObjs[i];
	if (!(mv.position() instanceof AmbulanceTeam)  &&  !mv.hasReachedDestination())
	  mv.setMotionlessObstructionAndMovingLaneList();
      }

      for (int i = mvObjs.length - 1;  i >= 0;  i --) {
	MovingObject mv = mvObjs[i];
	if (!(mv.position() instanceof AmbulanceTeam)  &&  !mv.hasReachedDestination())
	  mv.setMovingObstruction();
      }

      sortByObstruction();

      Iterator it;
      it = m_noWaitingNoChageList.iterator();
      while (it.hasNext())
	((MovingObject) it.next()).move();
      m_noWaitingNoChageList.clear();

      Collections.shuffle(m_waitingNoChangeList, RANDOM);
      it = m_waitingNoChangeList.iterator();
      while (it.hasNext()) {
	m_loopOrigin = null;
	moveAfterForwardMvObj((MovingObject) it.next());
      }
      m_waitingNoChangeList.clear();

      Collections.shuffle(m_noWaitingList, RANDOM);
      it = m_noWaitingList.iterator();
      while (it.hasNext())
	((MovingObject) it.next()).move();
      m_noWaitingList.clear();

      it = m_waitingMap.values().iterator();
      while (it.hasNext())
	dispatch((ArrayList) it.next());
      m_waitingMap.clear();

      if (USE_VIEWER) VIEWER.repaint();
    }
  }

  private ArrayList m_noWaitingNoChageList = new ArrayList();
  private ArrayList m_noWaitingList        = new ArrayList();
  private HashMap m_waitingMap             = new HashMap();
  private ArrayList m_waitingNoChangeList  = new ArrayList();
  private MovingObject m_loopOrigin;

  private void sortByObstruction() {
    for (int i = WORLD.movingObjectArray().length - 1;  i >= 0;  i --) {
      MovingObject mv = WORLD.movingObjectArray()[i];
      if (mv.hasReachedDestination())
	continue;
      if (mv.position() instanceof AmbulanceTeam) {
	mv.incrementNumReached();
	continue;
      }
      mv.setMoved(false);
      if (mv.obstruction() == Obstruction.DUMMY_OBSTRUCTION) {
	m_noWaitingNoChageList.add(mv);
      } else if (   mv.obstruction() instanceof Destination
		 || mv.obstruction() instanceof Entrance) {
	m_noWaitingList.add(mv);
      } else if (   mv.obstruction() instanceof Blockade
		 || mv.obstruction() instanceof CuttingLanes) {
	ArrayList follows = (ArrayList) m_waitingMap.get(mv.obstruction().motionlessPosition());
	if (follows == null) {
	  follows = new ArrayList();
	  m_waitingMap.put(mv.obstruction().motionlessPosition(), follows);
	}
	follows.add(mv);
      } else if (mv.obstruction() instanceof MovingObject) {
	m_waitingNoChangeList.add(mv);
      } else {
	if(ASSERT)Util.assert(false);
      }
      if(ASSERT)Util.assert(mv.obstruction().minSafeDistance() <= MAX_MIN_SAFE_DISTANCE,
			    "Update MAX_MIN_SAFE_DISTANCE to", mv.obstruction().minSafeDistance());
    }
  }

  private void moveAfterForwardMvObj(MovingObject mv) {
    if (!(mv.obstruction() instanceof MovingObject)) {
      if(ASSERT)Util.assert(mv.moved());
      return;
    }
    MovingObject fwdMv = (MovingObject) mv.obstruction();
    if (!fwdMv.moved()  &&  fwdMv != m_loopOrigin) {
      if (m_loopOrigin == null)
	m_loopOrigin = mv;
      if (fwdMv.obstruction() instanceof MovingObject)
	moveAfterForwardMvObj(fwdMv);
    }
    mv.move();
  }

  private void dispatch(ArrayList follows) {
    Collections.shuffle(follows, RANDOM);
    Iterator it = follows.iterator();
    while (it.hasNext()) {
      MovingObject mv = (MovingObject) it.next();
      Obstruction former = mv.obstruction();
      if (!mv.moved()) {
	mv.move();
	if (mv.obstruction() != former)
	  break;
      }
    }
    while (it.hasNext())
      ((MovingObject) it.next()).setMoved(true);
  }

  private void loadUnload() {
    Collections.shuffle(WORLD.ambulanceTeamList, RANDOM);
    Iterator it = WORLD.ambulanceTeamList.iterator();
    while (it.hasNext()) {
      AmbulanceTeam ambulance = (AmbulanceTeam) it.next();
      MovingObject target = ambulance.loadingMvObj();
      if (target == null)
	continue;
      if (ambulance.doLoad()) {
	if (target.lastMovingTime() == WORLD.time()) {
	  ambulance.resetLoad();
	  System.out.println("amb couldn't load target because it's moving.");
	} else {
	  if (target.position() instanceof MotionlessObject) {
	    target.takeOnAmbulance(ambulance);
	    ambulance.load();
	  } else {
	    ambulance.resetLoad();
	    System.out.println("amb couldn't load target because it is not on MotionlessObject.");
	  }
	}
      } else if (ambulance.doUnload()) {
	if(ASSERT)Util.assert(target.position() == ambulance, "wrong command; amb isn't target.position.");
	target.takeOffAmbulance();
	ambulance.unload();
      }
    }
  }
}
